from yt.utilities.object_registries import output_type_registry

import yt_aspect  # NOQA


def test_frontend_registry():
    assert "ASPECTDataset" in output_type_registry
    assert "PVTUDataset" in output_type_registry



import os

import pytest
import unyt
import yt

import yt_aspect  # NOQA


def get_file_path_from_data_info(dataset_type, data_info_dict, dataset_name):
    ds_info = data_info_dict[dataset_type][dataset_name]
    final_folder = os.path.split(ds_info["archive_path"])[-1]
    fi_dir = os.path.join(
        data_info_dict["base_dir"],
        ds_info["archive_path"],
        final_folder,
        ds_info["relative_unpacked_path"],
    )
    full_file = os.path.join(fi_dir, ds_info["sample_file"])
    return full_file


@pytest.mark.parametrize(
    "dataset_type, dataset_name",
    (
        ("ASPECT", "cartesian_3D_nproc4"),
        ("ASPECT", "cartesian_3D_nproc1"),
        ("PVTU", "cleaned_aspect"),
    ),
)
def test_3d_aspect_load(pvtu_test_data, dataset_type, dataset_name, tmp_path):

    fi = get_file_path_from_data_info(dataset_type, pvtu_test_data, dataset_name)

    if os.path.isfile(fi) is False:
        pytest.skip(f"Could not locate {fi}")

    ds = yt.load(fi)
    assert ds.dataset_type == dataset_type.lower()
    assert ds.dimensionality == 3

    ad = ds.all_data()
    T = ad[("connect0", "T")][0]
    if dataset_type == "ASPECT":
        assert T.units == unyt.K

    _ = ad[("connect0", "velocity_x")][0]

    outfi = tmp_path / "test.png"
    slc = yt.SlicePlot(ds, "x", ("connect0", "T"))
    slc.save(outfi)
    assert os.path.isfile(outfi)



from .data_structures import (  # noqa: F401
    ASPECTDataset,
    ASPECTUnstructuredIndex,
    ASPECTUnstructuredMesh,
    PVTUDataset,
    PVTUIndex,
    PVTUMesh,
)
from .fields import ASPECTFieldInfo, PVTUFieldInfo  # noqa: F401
from .io import IOHandlerASPECT, IOHandlerPVTU  # noqa: F401

__version__ = "0.1.0"



import os

import numpy as np
import xmltodict
from yt.utilities.io_handler import BaseIOHandler

from .util import decode_binary, decode_piece, type_decider


class IOHandlerPVTU(BaseIOHandler):
    _particle_reader = False
    _dataset_type = "pvtu"
    _INDEX_OFFSET = 1
    _vector_fields = ("Coordinates", "velocity")
    _xyz_to_dim = {"x": 0, "y": 1, "z": 2}
    _vel_vectors = ["velocity_" + dim for dim in ["x", "y", "z"]]

    def __init__(self, ds):
        self.filename = ds.index_filename
        super().__init__(ds)
        self.node_fields = ds._get_nod_names()
        self.elem_fields = ds._get_elem_names()

    def _read_particle_coords(self, chunks, ptf):
        pass

    def _read_particle_fields(self, chunks, ptf, selector):
        pass

    def _read_fluid_selection(self, chunks, selector, fields, size):
        # This needs to allocate a set of arrays inside a dictionary, where the
        # keys are the (ftype, fname) tuples and the values are arrays that
        # have been masked using whatever selector method is appropriate.  The
        # dict gets returned at the end and it should be flat, with selected
        # data.  Note that if you're reading grid data, you might need to
        # special-case a grid selector object.

        # always select all since these are chunks of a single mesh...
        ftype_list = []
        rv = {}
        for field in fields:
            ftype, fname = field
            rv[field] = []
            ftype_list.append(ftype)

        for chunk_id, chunk in enumerate(chunks):
            mesh = chunk.objs[0]
            npc = mesh.nodes_per_cell
            el_counts = np.array(mesh.element_count)

            # for now, we have all file info in single chunk
            mesh_name = f"connect{chunk_id}"
            if "all" in ftype_list or mesh_name in ftype_list:
                # mask here is the **element** mask. i.e.,
                #     shape(mask) = (n_elments,)
                # rather than (n_elements, n_verts). These apply to all fields,
                # so pull them out here:
                mask = selector.fill_mesh_cell_mask(mesh)  # global mask
                if mask is not None:
                    # each mesh piece will pull the connectivity, mask values
                    # for the indices corresponding to each vtu file
                    for vtu_id, vtu_fi in enumerate(mesh.filenames):

                        # the element mask for this piece
                        element_offset_start = el_counts[0:vtu_id].sum()
                        element_offset_end = element_offset_start + el_counts[vtu_id]
                        vtu_mask = mask[element_offset_start:element_offset_end]

                        # load this vtu's part of the mesh
                        vtu_file = os.path.join(self.ds.data_dir, vtu_fi)
                        f2id = self.ds.parameters["field_to_piece_index"][vtu_file]
                        with open(vtu_file) as data:
                            xml = xmltodict.parse(data.read())
                            xmlPieces = xml["VTKFile"]["UnstructuredGrid"]["Piece"]

                        if type(xmlPieces) != list:
                            xmlPieces = [xmlPieces]

                        for field in fields:
                            ftype, fname = field

                            # check if we are dealing with a velocity component
                            vdim = -1
                            if fname in self._vel_vectors:
                                vdim = self._xyz_to_dim[fname.split("_")[-1]]
                                fname = "velocity"

                            vtu_field, vtu_conn1d = self._read_single_vtu_field(
                                xmlPieces, fname, f2id, vectordim=vdim
                            )
                            vtu_field = vtu_field[vtu_conn1d]
                            vtu_field = vtu_field.reshape((vtu_conn1d.size // npc, npc))
                            vtu_field = vtu_field[vtu_mask, :]
                            rv[field].append(vtu_field)

        for field in fields:
            rv[field] = np.concatenate(rv[field]).astype(np.float64)

        return rv

    def _read_single_vtu_field(
        self, xmlPieces, fieldname, field_to_piece_index, vectordim=-1, ndims=3
    ):
        vtu_data = []
        pieceoff = 0
        vtu_conns = []

        for piece_id in range(0, len(xmlPieces)):
            field_id = field_to_piece_index[piece_id][fieldname]
            xmlPiece = xmlPieces[piece_id]
            data_array = xmlPiece["PointData"]["DataArray"][field_id]
            names = data_array["@Name"]
            if names == fieldname:
                types = type_decider[data_array["@type"]]
                metadata, data_field = decode_binary(
                    data_array["#text"].encode(), dtype=types
                )
                if vectordim >= 0:
                    # extract the dimension of the vector we want
                    data_field = data_field.reshape((data_field.size // ndims, ndims))
                    data_field = data_field[:, vectordim]
                vtu_data.append(data_field.astype(np.float64))

            coords, conn, offsets, cell_types = decode_piece(xmlPieces[piece_id])
            vtu_conns.extend(conn + pieceoff)
            # the connectivity array can repeat index references to coordinates
            # update the offset number by the number of coords points:
            pieceoff = pieceoff + coords.shape[0]

        vtu_data = np.concatenate(vtu_data)
        vtu_conns = np.array(vtu_conns)

        return vtu_data, vtu_conns

    def _read_chunk_data(self, chunk, fields):
        # This reads the data from a single chunk, and is only used for
        # caching.
        pass


class IOHandlerASPECT(IOHandlerPVTU):
    _dataset_type = "aspect"



from .data_structures import ASPECTDataset  # noqa: F401; noqa: F401
from .data_structures import ASPECTUnstructuredIndex  # noqa: F401
from .data_structures import ASPECTUnstructuredMesh  # noqa: F401
from .fields import ASPECTFieldInfo  # noqa: F401
from .io import IOHandlerASPECT  # noqa: F401



try:
    from defusedexpat.pyexpat import ExpatError  # noqa: F401
except ImportError:
    from xml.parsers.expat import ExpatError  # noqa: F401
import base64
import zlib

import numpy as np

type_decider = {"Float32": "<f4", "Float64": "<f8"}


def decode_piece(xmlPiece):
    coord_type = type_decider[xmlPiece["Points"]["DataArray"]["@type"]]

    _, coords = decode_binary(
        xmlPiece["Points"]["DataArray"]["#text"].encode(), dtype=coord_type
    )
    _, conn = decode_binary(
        xmlPiece["Cells"]["DataArray"][0]["#text"].encode(), dtype="u4"
    )
    _, offsets = decode_binary(
        xmlPiece["Cells"]["DataArray"][1]["#text"].encode(), dtype="u4"
    )
    _, cell_types = decode_binary(
        xmlPiece["Cells"]["DataArray"][2]["#text"].encode(), dtype="u1"
    )

    coords = coords.reshape((coords.size // 3, 3))

    return coords, conn, offsets, cell_types


def decode_binary(blob, use_zlib=True, dtype="<f4"):
    split_location = blob.find(b"==") + 2
    first = base64.decodebytes(blob[:split_location])
    second = base64.decodebytes(blob[split_location:])
    if zlib:
        second = zlib.decompress(second)
    return np.frombuffer(first, dtype="<f4"), np.frombuffer(second, dtype=dtype)


def _recursive_key_check(dict_inst: dict, nested_key_list: list) -> bool:

    key_len = len(nested_key_list)
    if key_len == 0:
        raise ValueError("nested_key_list must have at least 1 element")

    c_key = nested_key_list[0]  # the current key
    if key_len == 1:
        # at the final level
        return c_key in dict_inst
    else:
        if c_key in dict_inst and isinstance(dict_inst[c_key], dict):
            # go deeper
            return _recursive_key_check(dict_inst[c_key], nested_key_list[1:])
        else:
            # either the key DNE or the next level is not another
            # dictionary, in which case the next keys cannot exist
            return False



import json
import os
import stat

import numpy as np
import xmltodict
from yt.data_objects.index_subobjects.unstructured_mesh import UnstructuredMesh
from yt.data_objects.static_output import Dataset
from yt.data_objects.unions import MeshUnion
from yt.funcs import setdefaultattr
from yt.geometry.unstructured_mesh_handler import UnstructuredIndex
from yt.utilities.logger import ytLogger as mylog

from .fields import ASPECTFieldInfo, PVTUFieldInfo
from .util import ExpatError, _recursive_key_check, decode_piece


class PVTUMesh(UnstructuredMesh):
    _type_name = "pvtu"
    _con_args = (
        "mesh_id",
        "filename",
        "filenames",
        "vtu_offsets",
        "connectivity_indices",
        "connectivity_coords",
        "nodes_per_cell",
    )

    def __init__(
        self,
        mesh_id,
        filename,
        filenames,
        node_offsets,
        element_count,
        connectivity_indices,
        connectivity_coords,
        nodes_per_cell,
        index,
    ):
        super().__init__(
            mesh_id, filename, connectivity_indices, connectivity_coords, index
        )
        self.filenames = filenames
        self.node_offsets = node_offsets
        self.element_count = element_count
        self.nodes_per_cell = nodes_per_cell

        # some error checking


class PVTUIndex(UnstructuredIndex):
    """
    for on-disk vtu files. assumes we have a single mesh split across vtu files.
    """

    _mesh_class = PVTUMesh

    def __init__(self, ds, dataset_type="pvtu"):
        super().__init__(ds, dataset_type)

        # for when the mesh pieces are treated as independent meshes
        self.global_connectivity = None
        self.global_coords = None

    def _initialize_mesh(self):
        (
            connectivity,
            coords,
            node_offsets,
            el_count,
            nodes_per_cell,
        ) = self.ds._read_pieces()
        self.meshes = []
        mesh_files = self.ds.parameters["vtu_files"]
        pvtu_file = self.ds.parameter_filename
        mesh_id = 0

        # single global mesh
        mesh = self._mesh_class(
            mesh_id,
            pvtu_file,
            mesh_files,
            node_offsets,
            el_count,
            connectivity,
            coords,
            nodes_per_cell,
            self,
        )
        self.meshes.append(mesh)
        self.mesh_union = MeshUnion("mesh_union", self.meshes)

    def _detect_output_fields(self):
        elem_names = self.dataset.parameters["elem_names"]
        node_names = self.dataset.parameters["nod_names"]
        fnames = elem_names + node_names

        # this is aspect-specific
        if "velocity" in fnames:
            fnames.remove("velocity")
            for dim in ["x", "y", "z"]:
                fnames.append("velocity_" + dim)

        self.field_list = []
        for i in range(0, len(self.meshes)):
            self.field_list += [("connect%d" % i, fname) for fname in fnames]
        self.field_list += [("all", fname) for fname in fnames]


class PVTUDataset(Dataset):
    _index_class = PVTUIndex
    _field_info_class = PVTUFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="pvtu",
        storage_filename=None,
        units_override=None,
    ):
        """

        A class used to represent a single timestep of a on-disk ASPECT dataset

        """
        self.data_dir = os.path.dirname(filename)
        super().__init__(filename, dataset_type, units_override=units_override)
        self.fluid_types += self._get_fluid_types()
        self.index_filename = filename
        self.storage_filename = storage_filename
        self.default_field = ("all", "T")  # also aspect specific?
        self.default_fluid_type = self.default_field[0]
        self.sidecar_info = self._read_sidecar()

    def _get_sidecar_file(self):
        # returns the sidecar filename
        param_file = os.path.basename(self.parameter_filename)
        return os.path.join(
            self.data_dir, "." + os.path.splitext(param_file)[0] + ".json"
        )

    def _read_sidecar(self):
        # reads in sidecar file, returns empty dict if no file.

        # initialize for this .pvtu if it is not there
        sidecar = self._get_sidecar_file()
        if os.path.isfile(sidecar):
            with open(sidecar) as shandle:
                return json.load(shandle)[0]
        else:
            return {}

    def _write_sidecar(self):
        # writes current sidecar_info to json file (overwrite everything)
        with open(self._get_sidecar_file(), "w") as shandle:
            json.dump([self.sidecar_info], shandle)

    def _set_code_unit_attributes(self):
        setdefaultattr(self, "length_unit", self.quan(1.0, "m"))
        setdefaultattr(self, "mass_unit", self.quan(1.0, "kg"))
        setdefaultattr(self, "time_unit", self.quan(1.0, "s"))

    def _setup_coordinate_handler(self):
        # ensure correct ordering of axes so plots aren't rotated (z should always be
        # on the vertical axis).
        super()._setup_coordinate_handler()
        self.coordinates._x_pairs = (("x", "y"), ("y", "x"), ("z", "x"))
        self.coordinates._y_pairs = (("x", "z"), ("y", "z"), ("z", "y"))

    def _parse_parameter_file(self):

        # store the top level pvtu info
        with open(self.parameter_filename) as pvtu_fi:
            self.parameters["pXML"] = xmltodict.parse(pvtu_fi.read())

        pieces = self.parameters["pXML"]["VTKFile"]["PUnstructuredGrid"]["Piece"]
        if not isinstance(pieces, list):
            pieces = [pieces]
        self.parameters["vtu_files"] = [piece["@Source"] for piece in pieces]

        # periodicity: setting to False. Info is not in the pvtu files. Might
        # be able to check original.prm file if it exists.
        self._periodicity = (False, False, False)

        self.unique_identifier = self._get_unique_identifier()
        self.dimensionality = int(
            self.parameters["pXML"]["VTKFile"]["PUnstructuredGrid"]["PPoints"][
                "PDataArray"
            ]["@NumberOfComponents"]
        )

        self.parameters["nod_names"] = self._get_nod_names()
        self.parameters["elem_names"] = self._get_elem_names()
        self.parameters["num_meshes"] = len(pieces)
        self.current_time = self._get_current_time()
        self._set_domain_dims()
        self._set_dummy_parameters()

    def _set_dummy_parameters(self):
        # These attributes don't really make sense for ASPECT data, but yt warns
        # if they are not present, so we set them to dummy values here.
        self.cosmological_simulation = 0
        self.current_redshift = 0
        self.omega_lambda = 0
        self.omega_matter = 0
        self.hubble_constant = 0
        self.refine_by = 0

    def _get_fluid_types(self):
        with open(self.parameter_filename) as pvtu_fi:
            pXML = xmltodict.parse(pvtu_fi.read())
        n_pieces = len(pXML["VTKFile"]["PUnstructuredGrid"]["Piece"])
        fluid_types = tuple([f"connect{pcid}" for pcid in range(0, n_pieces)])
        fluid_types += ("all",)
        return fluid_types

    def _get_unique_identifier(self):
        return int(os.stat(self.parameter_filename)[stat.ST_CTIME])

    def _get_current_time(self):
        # might do something here...
        return 0.0

    def _get_elem_names(self):
        """

        Returns the names of the element vars, if available.

        """
        # everything right now is nodal data...
        return []

    def _get_nod_names(self):
        """

        Returns the names of the node vars

        """
        fields = self.parameters["pXML"]["VTKFile"]["PUnstructuredGrid"]["PPointData"][
            "PDataArray"
        ]
        return [field["@Name"] for field in fields]

    def _add_piece_to_field_map(self, src_file, xmlPieces):
        """
        This constructs a look-up dictionary, field_to_piece_index:

        field_to_piece_index[vtufilename][piece_id][fieldname] --> piece index

        where piece index is the index of xmlPiece["PointData"]["DataArray"]
        for the desired field for a given vtu file and piece of the vtu file.
        Allows quick lookup of the field rather than looping through
        xmlPiece["PointData"]["DataArray"] until the desired field is found.
        Used in aspect.io._read_fluid_selection

        Parameters:
            src_file : str
                vtu filename
            xmlPieces : list
                list containing the pieces of the mesh for this vtu file

        stores field_to_piece_index in the self.parameters dictionary.
        """

        if "field_to_piece_index" not in self.parameters.keys():
            self.parameters["field_to_piece_index"] = {}

        if src_file not in self.parameters["field_to_piece_index"].keys():
            src_map = {}
            for piece_id in range(0, len(xmlPieces)):
                src_map[piece_id] = {}
                xmlPiece = xmlPieces[piece_id]
                for field_id, data_array in enumerate(
                    xmlPiece["PointData"]["DataArray"]
                ):
                    fname = data_array["@Name"]
                    src_map[piece_id][fname] = field_id

            self.parameters["field_to_piece_index"][src_file] = src_map

    def _read_pieces_from_single_vtu(self, src_file):
        with open(src_file) as data:
            xml = xmltodict.parse(data.read())
            xmlPieces = xml["VTKFile"]["UnstructuredGrid"]["Piece"]

        # initialize the containers for each piece of this vtu file. These get
        # combined into single arrays after reading each piece.
        conns = []  # connectivity (node --> index of coordinate arrays)
        alloffs = []  # element offsets
        x, y, z = [], [], []  # node coordinates
        n_p_cells = []  # nodes per cell of each piece
        pieceoff = 0  # connectivity index offset between pieces

        # do some piece sanitization and documentation
        if type(xmlPieces) != list:
            # handles the case where we have a single piece in this vtu file
            xmlPieces = [xmlPieces]
        self._add_piece_to_field_map(src_file, xmlPieces)

        # loop over pieces of this vtu, decode and add to containers
        for piece_id in range(0, len(xmlPieces)):
            coords, conn, offsets, cell_types = decode_piece(xmlPieces[piece_id])

            if len(np.unique(cell_types)) > 1:
                # check for multiple cell types within a single piece
                raise NotImplementedError(
                    f"multiple cell types in piece {piece_id} of vtu {src_file}. "
                    f"The {self.dataset_type} frontend can only handle single "
                    "cell types at present."
                )

            n_nodes = conn.size
            n_cells = int(xmlPieces[piece_id]["@NumberOfCells"])
            n_p_cells.append(n_nodes // n_cells)

            x.extend(coords[:, 0])
            y.extend(coords[:, 1])
            z.extend(coords[:, 2])
            alloffs.extend(offsets)
            conns.extend(conn + pieceoff)  # connectivity across pieces

            # the connectivity array can repeat index references to coordinates
            # so update the offset number by the number of coords points
            # rather than the number of elements in conn:
            pieceoff += coords.shape[0]

        # concatenate some of the containers
        # adjust if we have 2d
        if self.dimensionality == 2:
            coords = np.array(np.column_stack([x, y]))
        else:
            coords = np.array(np.column_stack([x, y, z]))
        alloffs = np.array(alloffs)
        conns = np.array(conns)

        # check that cell types are the same across the pieces. Not handling
        # multiple cell types right now as ASPECT generally does not mix
        # elements. May want to generalize this for wider vtu applicability...
        n_p_cell_vals = np.unique(n_p_cells)
        if len(n_p_cell_vals) > 1:
            raise NotImplementedError(
                f"{src_file} contains multiple cell types. The ASPECT frontend "
                f"can only handle a single cell type at present"
            )
        npc = n_p_cell_vals[0]

        conns = conns.reshape((conns.size // npc, npc))

        return coords, conns, alloffs, npc

    def _read_pieces(self):
        # reads coordinates and connectivity from pieces, returns the mesh
        # concatenated across all pieces.
        with open(self.parameter_filename) as pvtu_fi:
            pXML = xmltodict.parse(pvtu_fi.read())

        # the separate vtu files
        vtu_pieces = pXML["VTKFile"]["PUnstructuredGrid"]["Piece"]
        if not isinstance(vtu_pieces, list):
            vtu_pieces = [vtu_pieces]

        # containers that will be concatenated across vtu files
        conlist = []  # list of 2D connectivity arrays for each vtu
        coordlist = []  # global coordinates in each vtu
        node_offsets = []  # the offset within each vtu
        element_count = []  # total elements in each vtu
        nodes_per_cell = []  # nodes per cell for each vtu

        current_offset = 0  # the current NODE offset to global coordinate index
        for src in vtu_pieces:
            src_file = os.path.join(self.data_dir, src["@Source"])
            coord, con, all_offs, npc = self._read_pieces_from_single_vtu(src_file)
            nodes_per_cell.append(npc)

            node_offsets.append(current_offset)
            element_count.append(con.shape[0])
            con = con + current_offset  # offset to global
            conlist.append(con.astype("i8"))
            coordlist.append(coord.astype("f8"))

            # the connectivity array can repeat index references to coordinates
            # so update the offset number by the number of coords points
            # rather than the number of elements in conn:
            current_offset += coord.shape[0]

        # check that we have the same cell types across vtu files
        nodes_per_cell = np.unique(nodes_per_cell)
        if len(nodes_per_cell) > 1:
            raise NotImplementedError(
                "Found different cell types across vtu files, can only "
                "handle single cell types at present"
            )
        nodes_per_cell = nodes_per_cell[-1]
        if nodes_per_cell > 8:
            mylog.info(
                "Found higher order elements: most operations will only use first order components."
            )

        # concatenate across into a single mesh.
        coordlist = np.vstack(coordlist)
        conlist = np.vstack(conlist).astype("i8")
        return conlist, coordlist, node_offsets, element_count, nodes_per_cell

    def _set_domain_dims(self):
        """
        Loads the global boundaries for the domain edge

        """

        # check our sidecar file first:
        self.sidecar_info = self._read_sidecar()
        left_edge = self.sidecar_info.get("domain_left_edge", None)
        right_edge = self.sidecar_info.get("domain_right_edge", None)

        if not left_edge or not right_edge:
            _, coord, _, _, _ = self._read_pieces()
            # vtu always has a z, even for 2D
            left_edge = [coord[:, i].min() for i in range(3)]
            right_edge = [coord[:, i].max() for i in range(3)]
            self.sidecar_info["domain_left_edge"] = left_edge
            self.sidecar_info["domain_right_edge"] = right_edge
            self._write_sidecar()

        left_edge = np.array(left_edge)
        right_edge = np.array(right_edge)

        # domain_width not set yet, calculate it:
        dwidth = right_edge - left_edge
        has_dim = np.array([wid > 0.0 for wid in dwidth])
        self.dimensionality = has_dim.sum()

        # set up pseudo-3D for 2d datasets here
        for idim in range(self.dimensionality, 3):
            right_edge[idim] = 1.0
            left_edge[idim] = 0.0

        self.domain_left_edge = left_edge
        self.domain_right_edge = right_edge
        self.domain_dimensions = np.ones(self.dimensionality, "int32")

    @classmethod
    def _is_valid(self, datafile, *args, **kwargs):
        is_pvtu = False
        if datafile.split(".")[-1] == "pvtu":
            required_keys = ["VTKFile", "PUnstructuredGrid", "Piece"]
            try:
                with open(datafile) as pvtu_fi:
                    pXML = xmltodict.parse(pvtu_fi.read())
                    is_pvtu = _recursive_key_check(pXML, required_keys)
            except (OSError, ExpatError):
                return False

        return is_pvtu


class ASPECTUnstructuredMesh(PVTUMesh):
    _type_name = "aspect_unstructured_mesh"


class ASPECTUnstructuredIndex(PVTUIndex):

    _mesh_class = ASPECTUnstructuredMesh

    def __init__(self, ds, dataset_type="aspect"):
        super().__init__(ds, dataset_type)


class ASPECTDataset(PVTUDataset):
    _index_class = ASPECTUnstructuredIndex
    _field_info_class = ASPECTFieldInfo

    def __init__(
        self,
        filename,
        dataset_type="aspect",
        storage_filename=None,
        units_override=None,
    ):
        super().__init__(
            filename,
            dataset_type=dataset_type,
            storage_filename=storage_filename,
            units_override=units_override,
        )

    @classmethod
    def _is_valid(self, datafile, *args, **kwargs):

        # first check if we have a pvtu file
        is_pvtu = super()._is_valid(datafile, *args, **kwargs)
        if is_pvtu is False:
            return False

        # now look for aspect specifics
        try:
            # pvtu files are small (~4K), so just parse the whole file and
            # check for the deal.II string
            with open(datafile) as data:
                return "generated by the deal.II library" in data.read()
        except OSError:
            return False



from yt.fields.field_info_container import FieldInfoContainer

# We need to specify which fields we might have in our dataset.  The field info
# container subclass here will define which fields it knows about.  There are
# optionally methods on it that get called which can be subclassed.


class PVTUFieldInfo(FieldInfoContainer):
    pass


class ASPECTFieldInfo(PVTUFieldInfo):

    known_other_fields = (
        ("T", ("K", [], None)),
        ("p", ("Pa", [], None)),
        ("viscosity", ("Pa*s", [], None)),
        ("velocity_x", ("m/s", [], None)),
        ("velocity_y", ("m/s", [], None)),
        ("velocity_z", ("m/s", [], None)),
        ("strain_rate", ("1/s", [], None)),
        ("thermal_diffusivity", ("m**2/s", [], None)),
        ("thermal_conductivity", ("W/m/K", [], None)),
        ("density", ("kg/m/m/m", [], None)),
        ("elastic_shear_modulus", ("Pa", [], None)),
        ("noninitial_plastic_strain", ("", [], None)),
        ("plastic_strain", ("", [], None)),
        ("plastic_yielding", ("", [], None)),
        ("artificial_viscosity", ("", [], None)),
        ("current_cohesions", ("", [], None)),
        ("current_friction_angles", ("", [], None)),
    )

    known_particle_fields = (
        # Identical form to above
        # ( "name", ("units", ["fields", "to", "alias"], # "display_name")),
    )

    log_fields = ["viscosity", "strain_rate"]

    def __init__(self, ds, field_list):

        # add the stress tensors to the known fields
        for xx in ["xx", "xy", "xz", "yy", "yz", "zz"]:
            self.known_other_fields += (("shear_stress_" + xx, ("Pa", [], None)),)
            self.known_other_fields += (("stress_" + xx, ("Pa", [], None)),)

        super().__init__(ds, field_list)
        for name in self:
            if name not in self.log_fields:
                self[name].take_log = False

    def setup_fluid_fields(self):
        # Here we do anything that might need info about the dataset.
        # You can use self.alias, self.add_output_field and self.add_field .
        pass

    def setup_particle_fields(self, ptype):
        # This will get called for every particle type.
        pass
